"""added Budget and BudgetEntry tables

Revision ID: 30c576f3691
Revises: 409d2d73ca30
Create Date: 2014-11-20 22:49:37.015323

"""

# revision identifiers, used by Alembic.
revision = '30c576f3691'
down_revision = '409d2d73ca30'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'Budgets',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['id'], ['Entities.id'], ),
        sa.ForeignKeyConstraint(['project_id'], ['Projects.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'BudgetEntries',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('budget_id', sa.Integer(), nullable=True),
        sa.Column('amount', sa.Float(), nullable=True),
        sa.ForeignKeyConstraint(['budget_id'], ['Budgets.id'], ),
        sa.ForeignKeyConstraint(['id'], ['Entities.id'], ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('BudgetEntries')
    op.drop_table('Budgets')



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest
from stalker import PriceList, Good


class PriceListTestCase(unittest.TestCase):
    """tests the PriceList class
    """

    def setUp(self):
        """set the test up
        """
        self.kwargs = {
            'name': 'Test Price List',
        }

    def test_goods_argument_is_skipped(self):
        """testing if the goods attribute will be an empty list when the goods
        argument is skipped
        """
        p = PriceList(**self.kwargs)
        self.assertEqual(p.goods, [])

    def test_goods_argument_is_None(self):
        """testing if the goods attribute  will be an empty list when the goods
        argument is None
        """
        self.kwargs['goods'] = None
        p = PriceList(**self.kwargs)
        self.assertEqual(p.goods, [])

    def test_goods_attribute_is_None(self):
        """testing if a TypeError will be raised when the goods attribute 
        is set to None
        """
        g1 = Good(name='Test Good')
        self.kwargs['goods'] = [g1]
        p = PriceList(**self.kwargs)
        self.assertEqual(p.goods, [g1])

        with self.assertRaises(TypeError) as cm:
            p.goods = None

        self.assertEqual(
            str(cm.exception),
            'Incompatible collection type: None is not list-like'
        )

    def test_goods_argument_is_not_a_list(self):
        """testing if a TypeError will be raised when the goods argument value
        is not a list
        """
        self.kwargs['goods'] = 'this is not a list'
        with self.assertRaises(TypeError) as cm:
            PriceList(**self.kwargs)

        self.assertEqual(
            str(cm.exception),
            'Incompatible collection type: str is not list-like'
        )

    def test_goods_attribute_is_not_a_list(self):
        """testing if a TypeError will be raised when the goods attribute is
        set to a value other than a list
        """
        g1 = Good(name='Test Good')
        self.kwargs['goods'] = [g1]
        p = PriceList(**self.kwargs)
        with self.assertRaises(TypeError) as cm:
            p.goods = 'this is not a list'

        self.assertEqual(
            str(cm.exception),
            'Incompatible collection type: str is not list-like'
        )

    def test_goods_argument_is_a_list_of_objects_which_are_not_goods(self):
        """testing if a TypeError will be raised when the goods argument is not
        a list of Good instances
        """
        self.kwargs['goods'] = ['not', 1, 'good', 'instances']
        with self.assertRaises(TypeError) as cm:
            PriceList(**self.kwargs)

        self.assertEqual(
            str(cm.exception),
            'PriceList.goods should be a list of stalker.model.bugdet.Good '
            'instances, not str'
        )

    def test_good_attribute_is_a_list_of_objects_which_are_not_goods(self):
        """testing if a TypeError will be raised when the goods attribute is
        not a list of Good instances
        """
        p = PriceList(**self.kwargs)
        with self.assertRaises(TypeError) as cm:
            p.goods = ['not', 1, 'good', 'instances']

        self.assertEqual(
            str(cm.exception),
            'PriceList.goods should be a list of stalker.model.bugdet.Good '
            'instances, not str'
        )

    def test_good_argument_is_working_properly(self):
        """testing if the good argument value is properly passed to the good
        attribute
        """
        g1 = Good(name='Good1')
        g2 = Good(name='Good2')
        g3 = Good(name='Good3')
        test_value = [g1, g2, g3]
        self.kwargs['goods'] = test_value
        p = PriceList(**self.kwargs)
        self.assertEqual(p.goods, test_value)

    def test_good_attribute_is_working_properly(self):
        """testing if the good attribute value can be properly set
        """
        g1 = Good(name='Good1')
        g2 = Good(name='Good2')
        g3 = Good(name='Good3')
        test_value = [g1, g2, g3]
        p = PriceList(**self.kwargs)
        self.assertNotEqual(p.goods, test_value)
        p.goods = test_value
        self.assertEqual(p.goods, test_value)



"""added Roles

Revision ID: 856e70016b2
Revises: 30c576f3691
Create Date: 2014-11-26 00:25:29.543411

"""

# revision identifiers, used by Alembic.
revision = '856e70016b2'
down_revision = '30c576f3691'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'Roles',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['id'], ['Entities.id'], ),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'Client_Users',
        sa.Column('uid', sa.Integer(), nullable=False),
        sa.Column('cid', sa.Integer(), nullable=False),
        sa.Column('rid', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['cid'], ['Clients.id'], ),
        sa.ForeignKeyConstraint(['rid'], ['Roles.id'], ),
        sa.ForeignKeyConstraint(['uid'], ['Users.id'], ),
        sa.PrimaryKeyConstraint('uid', 'cid')
    )

    #
    # read Users.client_id and create Client_Users entries accordingly
    #

    op.rename_table('User_Groups', 'Group_Users')
    op.rename_table('User_Departments', 'Department_Users')

    op.add_column(
        u'Department_Users',
        sa.Column('rid', sa.Integer(), nullable=True)
    )
    op.add_column(
        u'Project_Users',
        sa.Column('rid', sa.Integer(), nullable=True)
    )

    op.drop_column(u'Departments', 'lead_id')
    op.drop_column(u'Projects', 'lead_id')
    op.drop_column(u'Users', 'company_id')


def downgrade():
    op.add_column(
        u'Projects',
        sa.Column('lead_id', sa.INTEGER(), nullable=True)
    )
    op.add_column(
        u'Departments',
        sa.Column('lead_id', sa.INTEGER(), nullable=True)
    )

    op.drop_column(u'Project_Users', 'rid')
    op.drop_column(u'Department_Users', 'rid')

    op.rename_table('Department_Users', 'User_Departments')
    op.rename_table('Group_Users', 'User_Groups')

    op.add_column(
        u'Users',
        sa.Column('company_id', sa.INTEGER(), nullable=True)
    )

    op.drop_table('Client_Users')

    op.drop_table('Roles')



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest

from sqlalchemy import Column, Integer, ForeignKey

from stalker import db
from stalker.exceptions import CircularDependencyError
from stalker.models.entity import SimpleEntity
from stalker.models.mixins import DAGMixin

import logging
logging.getLogger('stalker.models.studio').setLevel(logging.DEBUG)


class DAGMixinFooMixedInClass(SimpleEntity, DAGMixin):
    """a class which derives from another which has and __init__ already
    """
    __tablename__ = "DAGMixinFooMixedInClasses"
    __mapper_args__ = {"polymorphic_identity": "DAGMixinFooMixedInClass"}
    dagMixinFooMixedInClass_id = Column(
        "id",
        Integer,
        ForeignKey("SimpleEntities.id"),
        primary_key=True
    )
    __id_column__ = 'dagMixinFooMixedInClass_id'

    def __init__(self, **kwargs):
        super(DAGMixinFooMixedInClass, self).__init__(**kwargs)
        DAGMixin.__init__(self, **kwargs)


class DAGMixinTestCase(unittest.TestCase):
    """tests the DAGMixin class
    """

    def setUp(self):
        """set the test up
        """
        # db.setup({'sqlalchemy.url': 'sqlite:///:memory:'})

        self.kwargs = {
            'name': 'Test DAG Mixin'
        }

    def tearDown(self):
        """clean up the test
        """
        # DBSession.remove()

    def test_parent_argument_is_skipped(self):
        """testing if the parent attribute will be None if the parent argument
        is skipped
        """
        d = DAGMixinFooMixedInClass(**self.kwargs)
        self.assertTrue(d.parent is None)

    def test_parent_argument_is_None(self):
        """testing if the parent attribute will be None if the parent argument
        is None
        """
        self.kwargs['parent'] = None
        d = DAGMixinFooMixedInClass(**self.kwargs)
        self.assertTrue(d.parent is None)

    def test_parent_argument_is_not_a_correct_class_instance(self):
        """testing if a TypeError will be raised if the parent argument is not
        in correct class
        """
        self.kwargs['parent'] = 'not a correct type'
        with self.assertRaises(TypeError) as cm:
            d = DAGMixinFooMixedInClass(**self.kwargs)

        self.assertEqual(
            str(cm.exception),
            'DAGMixinFooMixedInClass.parent should be an instance of '
            'DAGMixinFooMixedInClass class or derivative, not str'
        )

    def test_parent_attribute_is_not_a_correct_class_instance(self):
        """testing if a TypeError will be raised if the parent attribute is set
        to a wrong class instance
        """
        d = DAGMixinFooMixedInClass(**self.kwargs)
        with self.assertRaises(TypeError) as cm:
            d.parent = 'not a correct type'

        self.assertEqual(
            str(cm.exception),
            'DAGMixinFooMixedInClass.parent should be an instance of '
            'DAGMixinFooMixedInClass class or derivative, not str'
        )

    def test_parent_attribute_creates_a_cycle(self):
        """testing if a CircularDependency will be raised if a child is tried
        to be set as the parent.
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)

        self.kwargs['parent'] = d1
        d2 = DAGMixinFooMixedInClass(**self.kwargs)

        with self.assertRaises(CircularDependencyError) as cm:
            d1.parent = d2

        self.assertEqual(
            str(cm.exception),
            '\'<Test DAG Mixin (DAGMixinFooMixedInClass)> '
            '(DAGMixinFooMixedInClass) and '
            '<Test DAG Mixin (DAGMixinFooMixedInClass)> '
            '(DAGMixinFooMixedInClass) creates a circular dependency in their '
            '"children" attribute\''
        )

    def test_parent_argument_is_working_properly(self):
        """testing if the parent argument is working properly
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        self.kwargs['parent'] = d1

        d2 = DAGMixinFooMixedInClass(**self.kwargs)
        self.assertEqual(d1, d2.parent)

    def test_parent_attribute_is_working_properly(self):
        """testing if the parent attribute is working properly
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        d2 = DAGMixinFooMixedInClass(**self.kwargs)
        self.assertNotEqual(d2.parent, d1)
        d2.parent = d1
        self.assertEqual(d2.parent, d1)

    def test_children_attribute_is_an_empty_list_by_default(self):
        """testing if the children attribute is an empty list by default
        """
        d = DAGMixinFooMixedInClass(**self.kwargs)
        self.assertEqual(d.children, [])

    def test_children_attribute_is_set_to_None(self):
        """testing if a TypeError will be raised when the children attribute is
        set to None
        """
        d = DAGMixinFooMixedInClass(**self.kwargs)
        with self.assertRaises(TypeError) as cm:
            d.children = None

        self.assertEqual(
            str(cm.exception),
            'Incompatible collection type: None is not list-like'
        )

    def test_children_attribute_accepts_correct_class_instances_only(self):
        """testing if the children attribute accepts only correct class
        instances
        """
        d = DAGMixinFooMixedInClass(**self.kwargs)
        with self.assertRaises(TypeError) as cm:
            d.children = ['not', 1, '', 'of', 'correct', 'instances']

        self.assertEqual(
            str(cm.exception),
            'DAGMixinFooMixedInClass.children should be a list of '
            'DAGMixinFooMixedInClass (or derivative) instances, not str'
        )

    def test_children_attribute_is_working_properly(self):
        """testing if the children attribute is working properly
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        d2 = DAGMixinFooMixedInClass(**self.kwargs)
        d3 = DAGMixinFooMixedInClass(**self.kwargs)

        self.assertEqual(d1.children, [])
        d1.children.append(d2)
        self.assertEqual(d1.children, [d2])
        d1.children = [d3]
        self.assertEqual(d1.children, [d3])

    def test_is_leaf_attribute_is_read_only(self):
        """testing if the is_leaf attribute is a read only attribute
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        with self.assertRaises(AttributeError) as cm:
            setattr(d1, 'is_leaf', 'this will not work')

        self.assertEqual(
            str(cm.exception),
            "can't set attribute"
        )

    def test_is_leaf_attribute_is_working_properly(self):
        """testing if the is_leaf attribute is True for an instance without a
        child and False for another one with at least one child
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        d2 = DAGMixinFooMixedInClass(**self.kwargs)
        d3 = DAGMixinFooMixedInClass(**self.kwargs)
        d1.children = [d2, d3]
        self.assertFalse(d1.is_leaf)
        self.assertTrue(d2.is_leaf)
        self.assertTrue(d3.is_leaf)

    def test_is_root_attribute_is_read_only(self):
        """testing if the is_root attribute is a read only attribute
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        with self.assertRaises(AttributeError) as cm:
            setattr(d1, 'is_root', 'this will not work')

        self.assertEqual(
            str(cm.exception),
            "can't set attribute"
        )

    def test_is_root_attribute_is_working_properly(self):
        """testing if the is_root attribute is True for an instance without a
        parent and False for another instance with a parent
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        d2 = DAGMixinFooMixedInClass(**self.kwargs)
        d3 = DAGMixinFooMixedInClass(**self.kwargs)
        d1.children = [d2, d3]
        self.assertTrue(d1.is_root)
        self.assertFalse(d2.is_root)
        self.assertFalse(d3.is_root)

    def test_is_container_attribute_is_read_only(self):
        """testing if the is_container attribute is a read only attribute
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        with self.assertRaises(AttributeError) as cm:
            setattr(d1, 'is_container', 'this will not work')

        self.assertEqual(
            str(cm.exception),
            "can't set attribute"
        )

    def test_is_container_attribute_working_properly(self):
        """testing if the is_container attribute is True for an instance with
        at least one child and False for another with no child
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        d2 = DAGMixinFooMixedInClass(**self.kwargs)
        d3 = DAGMixinFooMixedInClass(**self.kwargs)
        d4 = DAGMixinFooMixedInClass(**self.kwargs)

        d1.children = [d2, d3]
        d2.children = [d4]
        self.assertTrue(d1.is_container)
        self.assertTrue(d2.is_container)
        self.assertFalse(d3.is_container)
        self.assertFalse(d4.is_container)

    def test_parents_property_is_read_only(self):
        """testing if the parents property is read-only
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        with self.assertRaises(AttributeError) as cm:
            setattr(d1, 'parents', 'this will not work')

        self.assertEqual(
            str(cm.exception),
            "can't set attribute"
        )

    def test_parents_property_is_working_properly(self):
        """testing if the parents property is read-only
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        d2 = DAGMixinFooMixedInClass(**self.kwargs)
        d3 = DAGMixinFooMixedInClass(**self.kwargs)
        d4 = DAGMixinFooMixedInClass(**self.kwargs)

        d1.children = [d2, d3]
        d2.children = [d4]

        self.assertEqual(d1.parents, [])
        self.assertEqual(d2.parents, [d1])
        self.assertEqual(d3.parents, [d1])
        self.assertEqual(d4.parents, [d1, d2])

    def test_walk_hierarchy_is_working_properly(self):
        """testing if the walk_hierarchy method is working properly
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        d2 = DAGMixinFooMixedInClass(**self.kwargs)
        d3 = DAGMixinFooMixedInClass(**self.kwargs)
        d4 = DAGMixinFooMixedInClass(**self.kwargs)

        d1.children = [d2, d3]
        d2.children = [d4]

        entities_walked = []
        for e in d1.walk_hierarchy():
            entities_walked.append(e)
        self.assertEqual(entities_walked, [d1, d2, d4, d3])

        entities_walked = []
        for e in d1.walk_hierarchy(method=1):
            entities_walked.append(e)
        self.assertEqual(entities_walked, [d1, d2, d3, d4])

        entities_walked = []
        for e in d2.walk_hierarchy():
            entities_walked.append(e)
        self.assertEqual(entities_walked, [d2, d4])

        entities_walked = []
        for e in d3.walk_hierarchy():
            entities_walked.append(e)
        self.assertEqual(entities_walked, [d3])

        entities_walked = []
        for e in d4.walk_hierarchy():
            entities_walked.append(e)
        self.assertEqual(entities_walked, [d4])


class DAGMixinDBTestCase(unittest.TestCase):
    """tests the DAGMixin class with a DB
    """

    def setUp(self):
        """set the test up
        """
        db.setup({'sqlalchemy.url': 'sqlite:///:memory:'})

        self.kwargs = {
            'name': 'Test DAG Mixin'
        }

    def tearDown(self):
        """clean up the test
        """
        db.DBSession.remove()

    def test_committing_data(self):
        """testing committing and retrieving data back
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        d2 = DAGMixinFooMixedInClass(**self.kwargs)
        d3 = DAGMixinFooMixedInClass(**self.kwargs)
        d4 = DAGMixinFooMixedInClass(**self.kwargs)

        d1.children = [d2, d3]
        d2.children = [d4]

        db.DBSession.add_all([d1, d2, d3, d4])
        db.DBSession.commit()

        del d1, d2, d3, d4

        all_data = DAGMixinFooMixedInClass.query.all()

        self.assertEqual(len(all_data), 4)
        self.assertTrue(isinstance(all_data[0], DAGMixinFooMixedInClass))
        self.assertTrue(isinstance(all_data[1], DAGMixinFooMixedInClass))
        self.assertTrue(isinstance(all_data[2], DAGMixinFooMixedInClass))
        self.assertTrue(isinstance(all_data[3], DAGMixinFooMixedInClass))

    def test_deleting_data(self):
        """testing deleting data
        """
        d1 = DAGMixinFooMixedInClass(**self.kwargs)
        d2 = DAGMixinFooMixedInClass(**self.kwargs)
        d3 = DAGMixinFooMixedInClass(**self.kwargs)
        d4 = DAGMixinFooMixedInClass(**self.kwargs)

        d1.children = [d2, d3]
        d2.children = [d4]

        db.DBSession.add_all([d1, d2, d3, d4])
        db.DBSession.commit()

        db.DBSession.delete(d1)
        db.DBSession.commit()

        all_data = DAGMixinFooMixedInClass.query.all()
        self.assertEqual(len(all_data), 0)



"""add PriceLists and Goods

Revision ID: 1c9c9c28c102
Revises: 856e70016b2
Create Date: 2015-01-26 13:05:50.050345

"""

# revision identifiers, used by Alembic.
revision = '1c9c9c28c102'
down_revision = '856e70016b2'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'PriceLists',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['id'], ['Entities.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'Goods',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('cost', sa.Float(), nullable=True),
        sa.Column('msrp', sa.Float(), nullable=True),
        sa.Column('unit', sa.String(length=64), nullable=True),
        sa.ForeignKeyConstraint(['id'], ['Entities.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'PriceList_Goods',
        sa.Column('price_list_id', sa.Integer(), nullable=False),
        sa.Column('good_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['good_id'], ['Goods.id'], ),
        sa.ForeignKeyConstraint(['price_list_id'], ['PriceLists.id'], ),
        sa.PrimaryKeyConstraint('price_list_id', 'good_id')
    )
    with op.batch_alter_table('BudgetEntries', schema=None) as batch_op:
        batch_op.add_column(sa.Column('cost', sa.Float(), nullable=True))
        batch_op.add_column(sa.Column('msrp', sa.Float(), nullable=True))
        batch_op.add_column(sa.Column('price', sa.Float(), nullable=True))
        batch_op.add_column(
            sa.Column('realized_total', sa.Float(), nullable=True))
        batch_op.add_column(
            sa.Column('unit', sa.String(length=64), nullable=True))

    with op.batch_alter_table('Budgets', schema=None) as batch_op:
        batch_op.add_column(
            sa.Column('parent_id', sa.Integer(), nullable=True))


def downgrade():
    with op.batch_alter_table('Budgets', schema=None) as batch_op:
        batch_op.drop_column('parent_id')

    with op.batch_alter_table('BudgetEntries', schema=None) as batch_op:
        batch_op.drop_column('unit')
        batch_op.drop_column('realized_total')
        batch_op.drop_column('price')
        batch_op.drop_column('msrp')
        batch_op.drop_column('cost')

    op.drop_table('PriceList_Goods')
    op.drop_table('Goods')
    op.drop_table('PriceLists')



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from sqlalchemy import Column, Integer, ForeignKey, Float, String, Table
from sqlalchemy.orm import relationship, validates
from stalker.db import Base

from stalker.models.entity import Entity
from stalker.models.mixins import ProjectMixin, DAGMixin


class Good(Entity):
    """Manages commercial items that is served by the Studio.

    A Studio can define service prices or items that's been sold by the Studio
    by using a list of commercial items.

    A Good has the following attributes

    :param msrp: The suggested retail price for this item.

    :param cost: The cost of this item to the Studio, so generally it is better
      to keep price of the related BudgetEntry bigger than this value to get
      profit by selling this item.

    :param unit: The unit of this item.
    """

    __auto_name__ = False
    __tablename__ = 'Goods'
    __mapper_args__ = {'polymorphic_identity': 'Good'}

    good_id = Column(
        'id',
        Integer,
        ForeignKey('Entities.id'),
        primary_key=True
    )

    price_lists = relationship(
        'PriceList',
        secondary='PriceList_Goods',
        primaryjoin='Goods.c.id==PriceList_Goods.c.good_id',
        secondaryjoin='PriceList_Goods.c.price_list_id==PriceLists.c.id',
        back_populates='goods',
        doc='PriceLists that this good is related to.'
    )

    cost = Column(Float, default=0.0)
    msrp = Column(Float, default=0.0)
    unit = Column(String(64))

    def __init__(self, cost=0.0, msrp=0.0, unit='', **kwargs):
        super(Good, self).__init__(**kwargs)
        self.cost = cost
        self.msrp = msrp
        self.unit = unit

    @validates('cost')
    def _validate_cost(self, key, cost):
        """validates the given cost value
        """
        if cost is None:
            cost = 0.0

        if not isinstance(cost, (float, int)):
            raise TypeError(
                '%s.cost should be a non-negative number, not %s' % (
                    self.__class__.__name__,
                    cost.__class__.__name__
                )
            )

        if cost < 0.0:
            raise ValueError(
                '%s.cost should be a non-negative number' %
                self.__class__.__name__
            )

        return cost

    @validates('msrp')
    def _validate_msrp(self, key, msrp):
        """validates the given msrp value
        """
        if msrp is None:
            msrp = 0.0

        if not isinstance(msrp, (float, int)):
            raise TypeError(
                '%s.msrp should be a non-negative number, not %s' % (
                    self.__class__.__name__,
                    msrp.__class__.__name__
                )
            )

        if msrp < 0.0:
            raise ValueError(
                '%s.msrp should be a non-negative number' %
                self.__class__.__name__
            )

        return msrp

    @validates('unit')
    def _validate_unit(self, key, unit):
        """validates the given unit value
        """
        if unit is None:
            unit = ''

        from stalker import __string_types__
        if not isinstance(unit, __string_types__):
            raise TypeError(
                '%s.unit should be a str, not %s' % (
                    self.__class__.__name__,
                    unit.__class__.__name__
                )
            )

        return unit


class PriceList(Entity):
    """Contains CommercialItems to create a list of items that is sold by the
    Studio.

    You can create different lists for items sold in this studio.
    """

    __auto_name__ = False
    __tablename__ = 'PriceLists'
    __mapper_args__ = {'polymorphic_identity': 'PriceList'}

    price_list_id = Column(
        'id',
        Integer,
        ForeignKey('Entities.id'),
        primary_key=True
    )

    goods = relationship(
        'Good',
        secondary='PriceList_Goods',
        primaryjoin='PriceLists.c.id==PriceList_Goods.c.price_list_id',
        secondaryjoin='PriceList_Goods.c.good_id==Goods.c.id',
        back_populates='price_lists',
        doc='Goods in this list.'
    )

    def __init__(self, goods=None, **kwargs):
        super(PriceList, self).__init__(**kwargs)
        if goods is None:
            goods = []
        self.goods = goods

    @validates('goods')
    def _validate_goods(self, key, good):
        """validates the given good value
        """
        from stalker import Good
        if not isinstance(good, Good):
            raise TypeError(
                '%s.goods should be a list of stalker.model.bugdet.Good '
                'instances, not %s' % (
                    self.__class__.__name__,
                    good.__class__.__name__
                )
            )

        return good

PriceList_Goods = Table(
    'PriceList_Goods', Base.metadata,
    Column(
        'price_list_id',
        Integer,
        ForeignKey('PriceLists.id'),
        primary_key=True
    ),
    Column(
        'good_id',
        Integer,
        ForeignKey('Goods.id'),
        primary_key=True
    )
)


class Budget(Entity, ProjectMixin, DAGMixin):
    """Manages project budgets

    Budgets manager :class:`.Project` budgets. You can create entries as
    instances of :class:`.BudgetEntry` class.
    """

    __auto_name__ = False
    __tablename__ = "Budgets"
    __mapper_args__ = {"polymorphic_identity": "Budget"}

    budget_id = Column(
        "id",
        Integer,
        ForeignKey("Entities.id"),
        primary_key=True
    )

    __id_column__ = 'budget_id'

    entries = relationship(
        'BudgetEntry',
        primaryjoin='BudgetEntries.c.budget_id==Budgets.c.id',
        uselist=True,
        cascade="all, delete-orphan"
    )

    def __init__(self, **kwargs):
        super(Budget, self).__init__(**kwargs)
        ProjectMixin.__init__(self, **kwargs)
        DAGMixin.__init__(self, **kwargs)

    @validates('entries')
    def _validate_entry(self, key, entry):
        """validates the given entry value
        """
        if not isinstance(entry, BudgetEntry):
            raise TypeError(
                '%(class)s.entries should be a list of BudgetEntry instances, '
                'not %(entry_class)s' % {
                    'class': self.__class__.__name__,
                    'entry_class': entry.__class__.__name__
                }
            )
        return entry


class BudgetEntry(Entity):
    """Manages budget entries in a budget

    With budget entries one can manage project budget costs. Each entry shows
    one component of a bigger budget.

    :param float cost: This should be the direct copy of the Good.cost that
      this item is related to.
    :param float msrp: Again this should be the direct copy of the Good.msrp
      that this item is related to.
    :param float price: The decided price of this entry. This is generally
      bigger than the cost and should be also bigger than msrp but the person
      that is editing the the budget which this entry is related to can decide
      to do a discount on this entry and give a different price. This attribute
      holds the proposed final price.
    :param float realized_total: This attribute is for holding the realized
      price of this entry. It can be the same number of the :attr:`.price`
      multiplied by the :attr:`.amount` or can be something else that reflects
      the reality. Generally it is for calculating the "service" cost/profit.
    :param str unit: The unit of this budget entry. Nothing so important for
      now. Just a text like "$/hour". Stalker doesn't do any unit conversions
      for now.
    :param float amount: Defines the amount of :class:`Good` that is in
      consideration for this entry.
    """

    __auto_name__ = True
    __tablename__ = "BudgetEntries"
    __mapper_args__ = {"polymorphic_identity": "BudgetEntry"}

    entry_id = Column(
        "id",
        Integer,
        ForeignKey("Entities.id"),
        primary_key=True
    )

    budget_id = Column(
        'budget_id',
        Integer,
        ForeignKey('Budgets.id')
    )

    budget = relationship(
        'Budget',
        primaryjoin='BudgetEntries.c.budget_id==Budgets.c.id',
        back_populates='entries',
        uselist=False
    )

    cost = Column(Float, default=0.0)
    msrp = Column(Float, default=0.0)

    price = Column(Float, default=0.0)
    realized_total = Column(Float, default=0.0)

    unit = Column(String(64))
    amount = Column(Float, default=0.0)

    def __init__(self,
                 budget=None,
                 cost=0,
                 msrp=0,
                 price=0,
                 realized_total=0,
                 unit='',
                 amount=0.0,
                 **kwargs):
        super(BudgetEntry, self).__init__(**kwargs)

        self.budget = budget
        self.cost = cost
        self.msrp = msrp

        self.price = price
        self.realized_total = realized_total

        self.amount = amount
        self.unit = unit

    @property
    def total(self):
        """returns the total cost of this entry
        """
        return self.amount * self.cost

    @validates('budget')
    def _validate_budget(self, key, budget):
        """validates the given budget value
        """
        if not isinstance(budget, Budget):
            raise TypeError(
                '%s.budget should be a Budget instance, not %s' % (
                    self.__class__.__name__, budget.__class__.__name__
                )
            )
        return budget

    @validates('cost')
    def _validate_cost(self, key, cost):
        """validates the given cost value
        """
        if cost is None:
            cost = 0.0

        if not isinstance(cost, (int, float)):
            raise TypeError(
                '%s.cost should be a number, not %s' % (
                    self.__class__.__name__, cost.__class__.__name__
                )
            )

        return float(cost)

    @validates('msrp')
    def _validate_msrp(self, key, msrp):
        """validates the given msrp value
        """
        if msrp is None:
            msrp = 0.0

        if not isinstance(msrp, (int, float)):
            raise TypeError(
                '%s.msrp should be a number, not %s' % (
                    self.__class__.__name__, msrp.__class__.__name__
                )
            )

        return float(msrp)

    @validates('price')
    def _validate_price(self, key, price):
        """validates the given price value
        """
        if price is None:
            price = 0.0

        if not isinstance(price, (int, float)):
            raise TypeError(
                '%s.price should be a number, not %s' % (
                    self.__class__.__name__, price.__class__.__name__
                )
            )

        return float(price)

    @validates('realized_total')
    def _validate_realized_total(self, key, realized_total):
        """validates the given realized_total value
        """
        if realized_total is None:
            realized_total = 0.0

        if not isinstance(realized_total, (int, float)):
            raise TypeError(
                '%s.realized_total should be a number, not %s' % (
                    self.__class__.__name__,
                    realized_total.__class__.__name__
                )
            )

        return float(realized_total)

    @validates('unit')
    def _validate_unit(self, key, unit):
        """validates the given unit value
        """
        if unit is None:
            unit = ''

        from stalker import __string_types__
        if not isinstance(unit, __string_types__):
            raise TypeError(
                '%s.unit should be a string, not %s' % (
                    self.__class__.__name__, unit.__class__.__name__
                )
            )

        return unit

    @validates('amount')
    def _validate_amount(self, key, amount):
        """validates the given amount value
        """
        if amount is None:
            amount = 0.0

        if not isinstance(amount, (int, float)):
            raise TypeError(
                '%s.amount should be a number, not %s' % (
                    self.__class__.__name__, amount.__class__.__name__
                )
            )

        return float(amount)



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2014 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest
from stalker.models.budget import Good


class GoodTestCase(unittest.TestCase):
    """tests the stalker.models.budget.Good class
    """

    def setUp(self):
        """set up the test
        """
        self.kwargs = {
            'name': 'Comp',
            'cost': 10,
            'msrp': 12,
            'unit': 'TL/hour'
        }

    def test_cost_argument_is_skipped(self):
        """testing if the cost attribute value will be 0.0 if the cost argument
        is skipped
        """
        self.kwargs.pop('cost')
        g = Good(**self.kwargs)
        self.assertEqual(g.cost, 0)

    def test_cost_argument_is_None(self):
        """testing if the cost attribute value will be 0.0 if the cost argument
        is None
        """
        self.kwargs['cost'] = None
        g = Good(**self.kwargs)
        self.assertEqual(g.cost, 0)

    def test_cost_attribute_is_None(self):
        """testing if the cost attribute will be 0.0 if it is set to None
        """
        g = Good(**self.kwargs)
        self.assertNotEqual(g.cost, 0)
        g.cost = None
        self.assertEqual(g.cost, 0)

    def test_cost_argument_is_not_a_number(self):
        """testing if a TypeError will be raised if cost argument is not a
        number
        """
        self.kwargs['cost'] = 'not a number'
        with self.assertRaises(TypeError) as cm:
            g = Good(**self.kwargs)

        self.assertEqual(
            str(cm.exception),
            'Good.cost should be a non-negative number, not str'
        )

    def test_cost_attribute_is_not_a_number(self):
        """testing if a TypeError will be raised if the cost attribute is set
        to something other than a number
        """
        g = Good(**self.kwargs)
        with self.assertRaises(TypeError) as cm:
            g.cost = 'not a number'

        self.assertEqual(
            str(cm.exception),
            'Good.cost should be a non-negative number, not str'
        )

    def test_cost_argument_is_zero(self):
        """testing if it is totally ok to set the cost to 0
        """
        self.kwargs['cost'] = 0
        g = Good(**self.kwargs)
        self.assertEqual(
            g.cost,
            0.0
        )

    def test_cost_attribute_is_zero(self):
        """testing if it is totally ok to test the cost attribute to 0
        """
        g = Good(**self.kwargs)
        self.assertNotEqual(g.cost, 0.0)
        g.cost = 0.0
        self.assertEqual(
            g.cost,
            0.0
        )

    def test_cost_argument_is_negative(self):
        """testing if a ValueError will be raised if the cost argument is a
        negative number
        """
        self.kwargs['cost'] = -10
        with self.assertRaises(ValueError) as cm:
            g = Good(**self.kwargs)

        self.assertEqual(
            str(cm.exception),
            'Good.cost should be a non-negative number'
        )

    def test_cost_attribute_is_negative(self):
        """testing if ValueError will be raised if the cost attribute is set to
        a negative number
        """
        g = Good(**self.kwargs)
        with self.assertRaises(ValueError) as cm:
            g.cost = -10

        self.assertEqual(
            str(cm.exception),
            'Good.cost should be a non-negative number'
        )

    def test_cost_argument_is_working_properly(self):
        """testing if the cost argument value is properly passed to the cost
        attribute
        """
        test_value = 113
        self.kwargs['cost'] = test_value
        g = Good(**self.kwargs)
        self.assertEqual(
            g.cost,
            test_value
        )

    def test_cost_attribute_is_working_properly(self):
        """testing if the cost attribute value can be properly changed
        """
        test_value = 145
        g = Good(**self.kwargs)
        self.assertNotEqual(
            g.cost,
            test_value
        )
        g.cost = test_value
        self.assertEqual(
            g.cost,
            test_value
        )

    def test_msrp_argument_is_skipped(self):
        """testing if the msrp attribute value will be 0.0 if the msrp argument
        is skipped
        """
        self.kwargs.pop('msrp')
        g = Good(**self.kwargs)
        self.assertEqual(g.msrp, 0)

    def test_msrp_argument_is_None(self):
        """testing if the msrp attribute value will be 0.0 if the msrp argument
        is None
        """
        self.kwargs['msrp'] = None
        g = Good(**self.kwargs)
        self.assertEqual(g.msrp, 0)

    def test_msrp_attribute_is_None(self):
        """testing if the msrp attribute will be 0.0 if it is set to None
        """
        g = Good(**self.kwargs)
        self.assertNotEqual(g.msrp, 0)
        g.msrp = None
        self.assertEqual(g.msrp, 0)

    def test_msrp_argument_is_not_a_number(self):
        """testing if a TypeError will be raised if msrp argument is not a
        number
        """
        self.kwargs['msrp'] = 'not a number'
        with self.assertRaises(TypeError) as cm:
            g = Good(**self.kwargs)

        self.assertEqual(
            str(cm.exception),
            'Good.msrp should be a non-negative number, not str'
        )

    def test_msrp_attribute_is_not_a_number(self):
        """testing if a TypeError will be raised if the msrp attribute is set
        to something other than a number
        """
        g = Good(**self.kwargs)
        with self.assertRaises(TypeError) as cm:
            g.msrp = 'not a number'

        self.assertEqual(
            str(cm.exception),
            'Good.msrp should be a non-negative number, not str'
        )

    def test_msrp_argument_is_zero(self):
        """testing if it is totally ok to set the msrp to 0
        """
        self.kwargs['msrp'] = 0
        g = Good(**self.kwargs)
        self.assertEqual(
            g.msrp,
            0.0
        )

    def test_msrp_attribute_is_zero(self):
        """testing if it is totally ok to test the msrp attribute to 0
        """
        g = Good(**self.kwargs)
        self.assertNotEqual(g.msrp, 0.0)
        g.msrp = 0.0
        self.assertEqual(
            g.msrp,
            0.0
        )

    def test_msrp_argument_is_negative(self):
        """testing if a ValueError will be raised if the msrp argument is a
        negative number
        """
        self.kwargs['msrp'] = -10
        with self.assertRaises(ValueError) as cm:
            g = Good(**self.kwargs)

        self.assertEqual(
            str(cm.exception),
            'Good.msrp should be a non-negative number'
        )

    def test_msrp_attribute_is_negative(self):
        """testing if ValueError will be raised if the msrp attribute is set to
        a negative number
        """
        g = Good(**self.kwargs)
        with self.assertRaises(ValueError) as cm:
            g.msrp = -10

        self.assertEqual(
            str(cm.exception),
            'Good.msrp should be a non-negative number'
        )

    def test_msrp_argument_is_working_properly(self):
        """testing if the msrp argument value is properly passed to the msrp
        attribute
        """
        test_value = 113
        self.kwargs['msrp'] = test_value
        g = Good(**self.kwargs)
        self.assertEqual(
            g.msrp,
            test_value
        )

    def test_msrp_attribute_is_working_properly(self):
        """testing if the msrp attribute value can be properly changed
        """
        test_value = 145
        g = Good(**self.kwargs)
        self.assertNotEqual(
            g.msrp,
            test_value
        )
        g.msrp = test_value
        self.assertEqual(
            g.msrp,
            test_value
        )

    def test_unit_argument_is_skipped(self):
        """testing if the unit attribute will be an empty string if the unit
        argument is skipped
        """
        self.kwargs.pop('unit')
        g = Good(**self.kwargs)
        self.assertEqual(g.unit, '')

    def test_unit_argument_is_None(self):
        """testing if the unit attribute will be an empty string if the unit
        argument is None
        """
        self.kwargs['unit'] = None
        g = Good(**self.kwargs)
        self.assertEqual(g.unit, '')

    def test_unit_attribute_is_set_to_None(self):
        """testing if the unit attribute will be an empty string if it is set
        to None
        """
        g = Good(**self.kwargs)
        self.assertNotEqual(g.unit, '')
        g.unit = None
        self.assertEqual(g.unit, '')

    def test_unit_argument_is_not_a_string(self):
        """testing if a TypeError will be raised if the unit argument is not a
        string
        """
        self.kwargs['unit'] = 12312
        with self.assertRaises(TypeError) as cm:
            g = Good(**self.kwargs)

        self.assertEqual(
            str(cm.exception),
            'Good.unit should be a str, not int'
        )

    def test_unit_attribute_is_not_a_string(self):
        """testing if a TypeError will be raised if the unit attribute is set
        to a value which is not a string
        """
        g = Good(**self.kwargs)
        with self.assertRaises(TypeError) as cm:
            g.unit = 2342

        self.assertEqual(
            str(cm.exception),
            'Good.unit should be a str, not int'
        )

    def test_unit_argument_is_working_properly(self):
        """testing if the unit argument value is properly passed to the unit
        attribute
        """
        test_value = 'this is my unit'
        self.kwargs['unit'] = test_value
        g = Good(**self.kwargs)
        self.assertEqual(g.unit, test_value)

    def test_unit_attribute_is_working_properly(self):
        """testing if the unit attribute value can be changed properly
        """
        test_value = 'this is my unit'
        g = Good(**self.kwargs)
        self.assertNotEqual(g.unit, test_value)
        g.unit = test_value
        self.assertEqual(g.unit, test_value)
